﻿using DicomObjects;
using DicomObjects.Enums;
using DicomObjects.UIDs;
using DicomObjects.Validation;
using Microsoft.Extensions.Logging;
using System;
using System.Collections.Generic;
using System.Drawing;

namespace PrintSCP
{
    class Program
    {
        #region Constants
        private const string LOGURL = "C:\\dicom log files";
        private const string Manufacturer = "Medical Connections";
        private const string ManufacturerModelName = "Demo Printer SCP";
        private const string PrinterStatus = "NORMAL";
        private const string DriverSerialNumber = "serial no 1234";
        private const string PrinterName = "PrinterName";
        private const string DateFormat = "yyyyMMdd";
        private const string TimeFormat = "hhmmss";
        private const string ServerStartingMessage = "Press any key to exit.\nServer Started, Listen Port ";
        private const string ServerStartingErrorMessage = "ERROR! Failed to Start Server : ";
        private const string FontTimesNewRoman = "Times New Roman";
        private const string FontConsolas = "Consolas";
        private const string FontArial = "Arial";
        private const string AssociationRequestFromMessage = "Association Request from";
        private const string Association = "Association";
        private const string ClosedText = "Closed";
        private const string Portrait = "PORTRAIT";
        private const string StartingPrinting = "Starting Printing";
        private const string PrintingFatalError = "Fatal Error! ";
        private const string PrintingFinished = "Printing Finished";
        #endregion


        public static DicomServer PrintServer { get; private set; }

        public static DicomDataSet PrinterObject { get; set; }
        readonly static DicomDataSetCollection BasicAnnotationBoxes = new DicomDataSetCollection();
        readonly static List<DisplayFormats> SupportedDisplayFormats = new List<DisplayFormats>();

        #region Main
        static void Main()
        {
            using var loggerFactory = LoggerFactory.Create(builder =>
            {
                builder
                    .AddConsole();
            });
            var logger = loggerFactory.CreateLogger<Program>();

            DicomGlobal.LogToLogger(logger, 0x3f);

            int port = 104;
            PrintServer = new DicomServer();
            try
            {
                PrintServer.Listen(port);
                PrintServer.NewAssociationType = typeof(PrinterAssociation);
                Log(ServerStartingMessage + port.ToString());
                PrintServer.AssociationRequest += PrintServer_AssociationRequest;
                PrintServer.AssociationClosed += PrintServer_AssociationClosed;
                PrintServer.NormalisedReceived += PrintServer_NormalisedReceived;
                PrintServer.VerifyReceived += (o, ee) => { ee.Status = 0; };
            }
            catch (DicomException ex)
            {
                Log(ServerStartingErrorMessage + ex.Message);
            }
            // Make a Global Printer Object
            MakePrinterDataset();

            initializeSupportedDisplayFomats();
            
            Console.ReadKey();
        }
        #endregion

        #region Event handlers
        private static void PrintServer_AssociationRequest(object Sender, DicomObjects.EventArguments.AssociationRequestArgs e)
        {
            Log($"{AssociationRequestFromMessage} {e.Association.CallingAET} At {e.Association.RemoteIP}");
            foreach (DicomContext ct in e.Contexts)
            {
                if (ct.AbstractSyntax == MetaSOPClasses.BasicGrayscalePrint)
                {
                    DicomDataSetCollection nullDataSetCollection = new DicomDataSetCollection();
                    var myPrinterDoc = new MyPrintDoc(nullDataSetCollection)
                    {
                        CurrentPageNumber = 0,
                        TotalPageNumber = 0
                    };
                    myPrinterDoc.PrintPage += MyPrinterDoc_PrintPage;
                    myPrinterDoc.EndPrint += MyPrinterDoc_EndPrint;
                    e.Association.Tag = myPrinterDoc;
                }
            }
        }

        private static void PrintServer_AssociationClosed(object Sender, DicomObjects.EventArguments.AssociationClosedArgs e)
        {
            Log($"{Association} {e.Association.AssociationNumber} {ClosedText}");
        }

        private static void PrintServer_NormalisedReceived(object Sender, DicomObjects.EventArguments.NormalisedReceivedArgs e)
        {
            PrinterAssociation assoc = (PrinterAssociation)e.RequestAssociation;
            assoc.MyPrinterDoc = e.RequestAssociation.Tag as MyPrintDoc;

            DicomDataSetCollection DataSets = assoc.DataSets;
            DicomDataSet filmBox = null;
            DicomDataSetCollection ImageBoxes;
            DicomDataSet imageBox;
            DicomDataSet command = e.Command;

            int operation = Convert.ToInt32(command[Keyword.CommandField].Value);
            string requestedSOPClass = command[Keyword.RequestedSOPClassUID].Value + "";
            string requestedSOPInstanceUID = command[Keyword.RequestedSOPInstanceUID].Value + "";
            string affectedSOPClass = command[Keyword.AffectedSOPClassUID].Value + "";
            string affectedSOPInstanceUID = command[Keyword.AffectedSOPInstanceUID].Value + "";

            switch (operation)
            {
                case 0x110: // N-GET
                    filmBox = DataSets[requestedSOPInstanceUID];
                    e.ResultDataSet = filmBox;
                    e.Status = 0;
                    break;

                case 0x140: // N-CREATE
                    filmBox = e.Request.Clone() as DicomDataSet; // Set Default Values					
                    filmBox.Add(Keyword.SOPClassUID, affectedSOPClass);

                    if (affectedSOPInstanceUID == "")
                        affectedSOPInstanceUID = DicomGlobal.NewUID();
                    filmBox.Add(Keyword.SOPInstanceUID, affectedSOPInstanceUID);

                    // FilmSession
                    if (affectedSOPClass == SOPClasses.BasicFilmSession)
                    {
                        assoc.FilmSession = new SessionDataSet();
                        foreach (DicomAttribute attr in filmBox)
                        {
                            assoc.FilmSession.Add(attr.Group, attr.Element, attr.Value);
                        }

                        assoc.FilmSession.ValidationOptions = ValidationOptions.None;

                        DataSets.Add(assoc.FilmSession);
                    }

                    // Specific to FilmBox
                    if (affectedSOPClass == SOPClasses.BasicFilmBox)
                    {
                        DecodeFormat(filmBox[Keyword.ImageDisplayFormat].Value.ToString(), out int cols, out int rows);
                        ImageBoxes = new DicomDataSetCollection();

                        // e.PresentationContextId is only available in DicomObjects Version 5.7 and onwards
                        bool color = e.RequestAssociation.AgreedContexts[e.PresentationContextId].AbstractSyntax == MetaSOPClasses.BasicColorPrint.ToString();

                        for (int i = 1; i <= rows * cols; i++)
                        {
                            imageBox = new DicomDataSet
                            {
                                InstanceUID = DicomGlobal.NewUID()
                            };
                            if (!color)
                                imageBox.Add(Keyword.ReferencedSOPClassUID, SOPClasses.BasicGrayscaleImageBox);
                            else
                                imageBox.Add(Keyword.ReferencedSOPClassUID, SOPClasses.BasicColorImageBox);

                            imageBox.Add(Keyword.ReferencedSOPInstanceUID, imageBox.InstanceUID);
                            DataSets.Add(imageBox);

                            ImageBoxes.Add(imageBox);
                        }

                        filmBox.Add(Keyword.ReferencedImageBoxSequence, ImageBoxes); // Referenced Image Box Sequence

                        //  Check if SOP class has been proposed, only then add it.
                        foreach (var agreedContext in e.RequestAssociation.AgreedContexts)
                            if (agreedContext.AbstractSyntax == SOPClasses.BasicAnnotationBox.ToString())
                                filmBox.Add(Keyword.ReferencedBasicAnnotationBoxSequence, BasicAnnotationBoxes);

                        // Also Add FilmBox to current session
                        assoc.FilmSession.FilmBoxes.Add(filmBox);

                        DataSets.Add(filmBox);
                    }

                    e.ResultDataSet = filmBox;
                    e.Status = 0;
                    break;

                case 0x120: // N-SET
                    foreach (DicomDataSet ba in BasicAnnotationBoxes)
                        if (ba[Keyword.ReferencedSOPInstanceUID].Value.ToString() == requestedSOPInstanceUID)
                            filmBox = ba;
                    if (filmBox == null)
                        filmBox = DataSets[requestedSOPInstanceUID];
                    //else
                    //    GenerateAnnotationLabels(e.Request);
                    foreach (DicomAttribute a in e.Request)
                    {
                        filmBox.Add(a.Group, a.Element, a.Value);
                    }
                    e.Status = 0;
                    break;

                case 0x130: // N-ACTION					

                    if (requestedSOPClass == SOPClasses.BasicFilmBox)       //  BasicFilmBox  is provided in the request
                    {
                        //  Increment pages
                        assoc.MyPrinterDoc.CurrentPageNumber += 1;

                        e.Status = PrintFilmBox(assoc, requestedSOPInstanceUID, DataSets);
                    }
                    else //  We assume that BasicFilmSession is provided instead
                    {
                        int status = 0xC000;

                        //  Iterate and Print each FilmBox
                        foreach (DicomDataSet filmbox in assoc.FilmSession.FilmBoxes)
                        {
                            //  Set total pages since we have the complete session details
                            assoc.MyPrinterDoc.TotalPageNumber = assoc.FilmSession.FilmBoxes.Count;
                            //  And print 1 at a time
                            assoc.MyPrinterDoc.CurrentPageNumber = 1;

                            status = PrintFilmBox(assoc, filmbox.InstanceUID, DataSets);
                        }
                        e.Status = status;
                    }
                    break;

                case 0x150: //N-Delete
                    DataSets.Remove(requestedSOPInstanceUID);

                    if (requestedSOPClass == SOPClasses.BasicFilmBox)
                        assoc.FilmSession.FilmBoxes.Remove(requestedSOPInstanceUID);

                    if (requestedSOPClass == SOPClasses.BasicFilmSession)
                        assoc.FilmSession = null;

                    e.Status = 0;
                    break;
            }
        }

        #region Printer event handlers
        private static void MyPrinterDoc_PrintPage(object sender, System.Drawing.Printing.PrintPageEventArgs e)
        {
            Log(StartingPrinting);

            MyPrintDoc mp = (MyPrintDoc)sender;

            DicomDataSetCollection datasets = mp.PrintDataSets;
            DicomDataSet filmbox = datasets[mp.UID];
            DicomDataSetCollection imageBoxes = filmbox[Keyword.ReferencedImageBoxSequence].Value as DicomDataSetCollection;


            DecodeFormat(filmbox[Keyword.ImageDisplayFormat].Value.ToString(), out int columns, out int rows);

            int margin = 5;
            int width, height;
            int paperWidth, paperHeight;

            if (!mp.DefaultPageSettings.Landscape)
            {
                paperWidth = mp.PrinterSettings.PaperSizes[0].Width;
                paperHeight = mp.PrinterSettings.PaperSizes[0].Height;
            }
            else
            {
                paperWidth = mp.PrinterSettings.PaperSizes[0].Height;
                paperHeight = mp.PrinterSettings.PaperSizes[0].Width;
            }

            width = ((paperWidth - margin) / columns) - margin;
            height = ((paperHeight - margin) / rows) - margin;

            Point LeftTop = new Point();
            Font textFont = new Font("Arial", 8);
            string pageNumber = "Page " + mp.CurrentPageNumber.ToString();
            e.Graphics.DrawString(pageNumber, textFont, Brushes.Blue, 4, 4);

            for (int row = 0; row < rows; row++)
            {
                for (int column = 0; column < columns; column++)
                {
                    DicomDataSet ds2 = datasets[imageBoxes[row * columns + column][Keyword.ReferencedSOPInstanceUID].Value.ToString()];
                    DicomAttribute imageAtt;

                    // if Mono Image
                    if (ds2[Keyword.BasicGrayscaleImageSequence].Exists)
                    {
                        imageAtt = ds2[Keyword.BasicGrayscaleImageSequence];
                    }
                    // if not mono image, then check for a color image
                    else
                    {
                        imageAtt = ds2[Keyword.BasicColorImageSequence];
                    }

                    if (imageAtt.Exists) // has real image then print
                    {
                        DicomDataSetCollection dss2 = (DicomDataSetCollection)imageAtt.Value;
                        // Get the DicomImage to print
                        DicomDataSet dcmImage = dss2[0];
                        // Set up the Print Area						 
                        int min;
                        int offSetX = 0;
                        int offSetY = 0;
                        if (width > height)
                        {
                            min = height;
                            offSetX = (paperWidth - min * columns - (columns + 1) * margin) / 2;
                        }
                        else
                        {
                            min = width;
                            offSetY = (paperHeight - min * rows - (rows + 1) * margin) / 2;
                        }

                        LeftTop.X = column * (min + margin) + margin + offSetX;
                        LeftTop.Y = row * (min + margin) + margin + offSetY;

                        Rectangle PrintArea = new Rectangle(LeftTop.X, LeftTop.Y, min, min);

                        //  Add Annotation Labels if any
                        //if (AnnotationLabels != null && AnnotationLabels.Count > 0)
                        //{
                        //    dcmImage.Labels = AnnotationLabels;
                        //    AnnotationLabels = null;
                        //}

                        // Get Bitmap Image by DicomImage.Bitmap() method
                        Image printImage = dcmImage.Bitmap();

                        try
                        {
                            // Print Image here
                            e.Graphics.DrawImage(printImage, PrintArea);
                            mp.Status = 0;
                        }
                        catch (Exception ex)
                        {
                            Log(PrintingFatalError + ex.Message);
                        }
                    }
                }
            }
        }

        private static void MyPrinterDoc_EndPrint(object sender, System.Drawing.Printing.PrintEventArgs e)
        {
            Log(PrintingFinished);
        }
        #endregion
        #endregion

        #region Utilities
        private static void MakePrinterDataset()
        {
            //  Dummy values as placeholder
            DicomDataSet dataSet = new DicomDataSet
            {
                InstanceUID = WellKnownInstances.Printer
            };
            dataSet.Add(Keyword.SOPClassUID, SOPClasses.Printer);
            dataSet.Add(Keyword.Manufacturer, Manufacturer);
            dataSet.Add(Keyword.ManufacturerModelName, ManufacturerModelName);
            dataSet.Add(Keyword.PrinterStatus, PrinterStatus);
            dataSet.Add(Keyword.PrinterStatusInfo, PrinterStatus);
            dataSet.Add(Keyword.DeviceSerialNumber, DriverSerialNumber);
            dataSet.Add(Keyword.SoftwareVersions, DicomGlobal.AssemblyVersion);
            dataSet.Add(Keyword.PrinterName, PrinterName);
            dataSet.Add(Keyword.DateOfLastCalibration, DateTime.Now.Date.ToString(DateFormat));
            dataSet.Add(Keyword.TimeOfLastCalibration, DateTime.Now.TimeOfDay.ToString(TimeFormat));
            PrinterObject = dataSet;
        }

        private static void initializeSupportedDisplayFomats()
        {
            //  Defining SCP supported Annotation boxes
            DicomDataSet annotation1 = new DicomDataSet();
            annotation1.Add(Keyword.ReferencedSOPClassUID, SOPClasses.BasicAnnotationBox);
            annotation1.Add(Keyword.ReferencedSOPInstanceUID, DicomGlobal.NewUID());
            BasicAnnotationBoxes.Add(annotation1);

            DicomDataSet annotation2 = new DicomDataSet();
            annotation2.Add(Keyword.ReferencedSOPClassUID, SOPClasses.BasicAnnotationBox);
            annotation2.Add(Keyword.ReferencedSOPInstanceUID, DicomGlobal.NewUID());
            BasicAnnotationBoxes.Add(annotation2);

            //  Defining display formats for each AnnotationID
            SupportedDisplayFormats.Add(new DisplayFormats(1, new RectangleF(10, 10, 100, 100), new Font(FontTimesNewRoman, 10)));
            SupportedDisplayFormats.Add(new DisplayFormats(2, new RectangleF(150, 50, 100, 200), new Font(FontConsolas, 20)));
            SupportedDisplayFormats.Add(new DisplayFormats(3, new RectangleF(80, 150, 100, 200), new Font(FontArial, 8)));
        }

        private static void DecodeFormat(string format, out int x, out int y)
        {
            int p, p1;
            p = format.IndexOf(@"\");
            p1 = format.IndexOf(",", p + 1);
            x = Convert.ToInt32(format.Substring(p + 1, p1 - p - 1));
            y = Convert.ToInt32(format.Substring(p1 + 1));
        }

        private static int PrintFilmBox(PrinterAssociation assoc, string requestedSOPInstanceUID, DicomDataSetCollection DataSets)
        {
            assoc.MyPrinterDoc.UID = requestedSOPInstanceUID;
            assoc.MyPrinterDoc.PrintDataSets = DataSets;

            DicomDataSet filmbox = DataSets[requestedSOPInstanceUID];
            string orientation = filmbox[Keyword.FilmOrientation].Value.ToString();
            if (orientation == Portrait)
            {
                assoc.MyPrinterDoc.DefaultPageSettings.Landscape = false;
            }
            else
            {
                assoc.MyPrinterDoc.DefaultPageSettings.Landscape = true;
            }

            //  Print Document
            assoc.MyPrinterDoc.Print();

            //  return Status
            return assoc.MyPrinterDoc.Status;
        }

        /* Disused: DicomLabel not implemented in Core */
        //private void GenerateAnnotationLabels(DicomDataSet requestDS)
        //{
        //    DisplayFormats chosenDisplayFormat = null;
        //    foreach (var df in SupportedDisplayFormats)
        //    {
        //        if (df.ID == int.Parse(requestDS[Keyword.AnnotationPosition].Value.ToString()))
        //            chosenDisplayFormat = df;

        //        if (chosenDisplayFormat != null)
        //        {
        //            DicomLabel l = new DicomLabel()
        //            {
        //                Text = requestDS[Keyword.TextString].Value.ToString(),
        //                Area = chosenDisplayFormat.Area,
        //                Font = chosenDisplayFormat.FontType
        //            };

        //            if (AnnotationLabels == null)
        //                AnnotationLabels = new DicomLabelCollection();
        //            AnnotationLabels.Add(l);
        //            chosenDisplayFormat = null;
        //        }
        //    }
        //}
        #endregion

        #region Logging
        private static void Log(string s)
        {
            Console.WriteLine(s);
        }
        #endregion
    }
}
